"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""

import os
import unittest
from sphinx.application import Sphinx
from sphinx.util.docutils import docutils_namespace
from sphinx.errors import SphinxWarning
from functools import wraps

from tests import tmpdir


def func_returns_int():
    return 1


def func_takes_arguments(arg1):  # pylint: disable=unused-argument
    return {}


non_callable_object = 1


class MockApp(Sphinx):
    """
    A subclass of :class:`Sphinx`
    """

    def __init__(self,):

        Sphinx.__init__(
            self,
            srcdir=os.path.abspath(""),
            confdir=os.path.abspath(""),
            outdir="result",
            doctreedir="doctreedir",
            buildername="html",
            warningiserror=True,
        )


class with_sphinx_app(object):
    def __init__(self, *sphinxargs, **sphinxkwargs):
        self.sphinxargs = sphinxargs
        self.sphinxkwargs = sphinxkwargs

    def __call__(self, func):
        @wraps(func)
        def decorator(*args, **kwargs):
            app = None
            sphinxkwargs = dict(self.sphinxkwargs)
            with docutils_namespace():
                app = MockApp(*self.sphinxargs, **sphinxkwargs)
                return func(*(args + (app,)), **kwargs)

        return decorator


class TestSphinxExtension(unittest.TestCase):
    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_invalid_setup(self, app):
        missing_func = """
        .. configsuite::
            :module: tests
        """

        with open("index.rst", "w") as f:
            f.write(missing_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue(
            "Expected 'module.function', got tests" in str(context.exception)
        )

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_invalid_module(self, app):
        invalid_func = """
        .. configsuite::
            :module: tests.non_existing_module.non_existing_function
        """

        with open("index.rst", "w") as f:
            f.write(invalid_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue("non_existing_module" in str(context.exception))

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_invalid_func(self, app):
        invalid_func = """
        .. configsuite::
            :module: tests.data.non_existing_function
        """

        with open("index.rst", "w") as f:
            f.write(invalid_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue(
            "has no attribute 'non_existing_function'" in str(context.exception)
        )

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_func_must_be_callable(self, app):
        invalid_func = """
        .. configsuite::
            :module: tests.test_sphinx_extension.non_callable_object
        """

        with open("index.rst", "w") as f:
            f.write(invalid_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue(
            "must provide a callable function, the 'non_callable_object' was not"
            in str(context.exception)
        )

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_func_takes_no_arguments(self, app):
        invalid_func = """
        .. configsuite::
            :module: tests.test_sphinx_extension.func_takes_arguments
        """

        with open("index.rst", "w") as f:
            f.write(invalid_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue(
            "not require any arguments, the 'func_takes_arguments' takes 1 argument(s)"
            in str(context.exception)
        )

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_func_invalid_return_type(self, app):
        invalid_func = """
        .. configsuite::
            :module: tests.test_sphinx_extension.func_returns_int
        """

        with open("index.rst", "w") as f:
            f.write(invalid_func)

        with self.assertRaises(SphinxWarning) as context:
            app.build()

        self.assertTrue(
            "The function 'func_returns_int' returned" in str(context.exception)
        )

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_sphinx_build(self, app):
        app.build()
        with open(os.path.join(app.outdir, "index.html"), "r") as f:
            html = f.read()

        contents = [
            "Text above",
            "The PetMeet configuration is to give an",
            "name",
            "Your full name.",
            "Text below",
        ]

        for content in contents:
            self.assertTrue(content in html)

    @tmpdir("tests/data/doc")
    @with_sphinx_app()
    def test_sphinx_complete_generation(self, app):
        complete_schema = """
        .. configsuite::
            :module: tests.data.car.build_schema_with_validators_and_transformators
        """
        with open("index.rst", "w") as f:
            f.write(complete_schema)

        app.build()
        with open(os.path.join(app.outdir, "index.html"), "r") as f:
            html = f.read()

        contents = [
            "production_date",
            "country",
            #  "Norway",
            "tire",
            "dimension",
            # "17",
            "owner",
            "location",
            "incidents",
            "Convert to cm",
            "Convert to cm - ignoring context",
            "Is x a valid date",
            "Is x a valid dimension",
        ]
        for content in contents:
            self.assertTrue(content in html)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


class ValidationError(object):
    def __init__(self, message, key_path, layer=None):
        self._message = message
        self._key_path = key_path
        self._layer = layer

    def __repr__(self):
        return "{}(msg={}, key_path={}, layer={})".format(
            self.__class__.__name__, self.msg, self.key_path, self.layer
        )

    @property
    def msg(self):
        return self._message

    @property
    def key_path(self):
        return self._key_path

    @property
    def layer(self):
        return self._layer

    def create_layer_error(self, layer):
        return self.__class__(self.msg, self.key_path, layer=layer)

    def __eq__(self, other):
        return (
            self.msg == other.msg
            and self.key_path == other.key_path
            and self.layer == other.layer
        )

    def __neq__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.msg, self.key_path, self.layer))


class UnknownKeyError(ValidationError):
    pass


class MissingKeyError(ValidationError):
    pass


class InvalidValueError(ValidationError):
    pass


class InvalidTypeError(ValidationError):
    pass


class TransformationError(ValidationError):
    pass


class ContextExtractionError(ValidationError):
    pass



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import datetime

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


def build_schema():
    return {
        MK.Type: types.NamedDict,
        MK.Content: {
            "production_date": {MK.Type: types.DateTime},
            "country": {
                MK.Type: types.String,
                MK.Default: "Norway",
                MK.Required: False,
                MK.AllowNone: True,
            },
            "tire": {
                MK.Type: types.NamedDict,
                MK.Content: {
                    "dimension": {
                        MK.Type: types.Number,
                        MK.Default: 17,
                        MK.Required: False,
                        MK.AllowNone: True,
                    },
                    "rim": {
                        MK.Type: types.String,
                        MK.Default: "green",
                        MK.Required: False,
                        MK.AllowNone: True,
                    },
                },
            },
            "owner": {
                MK.Type: types.Dict,
                MK.Content: {
                    MK.Key: {MK.Type: types.String},
                    MK.Value: {
                        MK.Type: types.NamedDict,
                        MK.Content: {
                            "name": {MK.Type: types.String},
                            "location": {
                                MK.Type: types.String,
                                MK.Default: "Earth",
                                MK.Required: False,
                                MK.AllowNone: True,
                            },
                        },
                    },
                },
            },
            "incidents": {
                MK.Type: types.List,
                MK.Content: {
                    MK.Item: {
                        MK.Type: types.NamedDict,
                        MK.Content: {
                            "location": {MK.Type: types.String},
                            "casualties": {
                                MK.Type: types.Number,
                                MK.Default: 0,
                                MK.Required: False,
                                MK.AllowNone: True,
                            },
                        },
                    }
                },
            },
        },
    }


def build_schema_with_validators_and_transformators():
    schema = build_schema()
    dimension_schema = schema[MK.Content]["tire"][MK.Content]["dimension"]
    dimension_schema[MK.Transformation] = inch_to_cm
    dimension_schema[MK.ContextTransformation] = inch_to_cm_context_based

    dimension_schema[MK.ElementValidators] = [is_valid_dimension]
    date_content = {
        MK.Type: types.DateTime,
        MK.Default: datetime.datetime(1999, 1, 1),
        MK.ContextValidators: [is_valid_date],
        MK.Required: False,
        MK.AllowNone: True,
    }

    owner_schema = schema[MK.Content]["owner"]
    owner_schema[MK.Content][MK.Value][MK.Content]["date"] = date_content
    return schema


@configsuite.validator_msg("Is x a valid dimension")
def is_valid_dimension(dimension):
    return 30 <= dimension <= 50


@configsuite.validator_msg("Is x a valid date")
def is_valid_date(date, context):
    return date >= context.production_date


@configsuite.transformation_msg("Convert to cm")
def inch_to_cm(l):
    return l * 2.54


@configsuite.transformation_msg("Convert to cm - ignoring context")
def inch_to_cm_context_based(l, _):
    return l * 2.54


def extract_production_date(snapshot):
    Context = collections.namedtuple("Context", ("production_date",))
    return Context(production_date=snapshot.production_date)


def build_all_default_config():
    return {
        "production_date": datetime.datetime(2000, 1, 1),
        "owner": {},
        "incidents": [],
    }


def build_config():
    return {
        "production_date": datetime.datetime(2000, 1, 1),
        "tire": {"dimension": 15},
        "owner": {
            "first entry": {"name": "Johan", "location": "Svalbard"},
            "second entry": {"name": "Svein"},
        },
        "incidents": [
            {"location": "whereabouts"},
            {"location": "somewhere else", "casualties": 1},
        ],
    }



"""Copyright 2020 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest
import warnings

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


class TestRequiredDeprecated(unittest.TestCase):
    def test_explicit_required_is_deprecated(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {"some_key": {MK.Type: types.String, MK.Required: True}},
        }
        with warnings.catch_warnings(record=True) as wc:
            configsuite.ConfigSuite({}, schema)
            self.assertEqual(1, len(wc))
            self.assertEqual(__file__, wc[0].filename)
            self.assertIn(
                "Use `ConfigSuite(..., deduce_required=True)`", str(wc[0].message)
            )

    def test_specifying_required_is_deprecated_when_deducing(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {"some_key": {MK.Type: types.String, MK.Required: True}},
        }
        with warnings.catch_warnings(record=True) as wc:
            configsuite.ConfigSuite({}, schema, deduce_required=True)
            self.assertEqual(1, len(wc))
            self.assertEqual(__file__, wc[0].filename)
            self.assertIn("Please remove them from your schema", str(wc[0].message))

    def test_no_deprecation_warning_when_deducing(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {"some_key": {MK.Type: types.String}},
        }
        with warnings.catch_warnings(record=True) as wc:
            suite = configsuite.ConfigSuite({}, schema, deduce_required=True)
            self.assertFalse(suite.valid)
            self.assertEqual(0, len(wc))

    def test_no_error_when_deducing_from_allow_none(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {"some_key": {MK.Type: types.String, MK.AllowNone: True}},
        }
        with warnings.catch_warnings(record=True) as wc:
            suite = configsuite.ConfigSuite({}, schema, deduce_required=True)
            self.assertTrue(suite.valid)
            self.assertEqual(0, len(wc))

    def test_no_error_when_deducing_from_default(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {"some_key": {MK.Type: types.String, MK.Default: "A string"}},
        }
        with warnings.catch_warnings(record=True) as wc:
            suite = configsuite.ConfigSuite({}, schema, deduce_required=True)
            self.assertTrue(suite.valid)
            self.assertEqual(0, len(wc))



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest

import configsuite
from configsuite import MetaKeys as MK
from .data import computers


class TestNotAllowNoneContainers(unittest.TestCase):
    def test_not_allow_allownone_for_list(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["externals"][MK.AllowNone] = False

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(config, schema)
        self.assertIn(
            "AllowNone can only be used for BasicType", str(error_context.exception)
        )

    def test_not_allow_allownone_for_dict(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["software"][MK.AllowNone] = False

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(config, schema)
        self.assertIn(
            "AllowNone can only be used for BasicType", str(error_context.exception)
        )

    def test_not_allow_allownone_for_named_dict(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["components"][MK.AllowNone] = False

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite(config, schema)
        self.assertIn(
            "AllowNone can only be used for BasicType", str(error_context.exception)
        )


class TestAllowNoneBasicType(unittest.TestCase):
    def test_basictype_explicit_none(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["OS"][MK.AllowNone] = True
        schema[MK.Content]["OS"][MK.Required] = False

        config["OS"] = None
        config_suite = configsuite.ConfigSuite(config, schema)

        self.assertTrue(config_suite.valid)
        self.assertIsNone(config_suite.snapshot.OS)

    def test_basictype_implicit_none(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["OS"][MK.AllowNone] = True
        schema[MK.Content]["OS"][MK.Required] = False
        config.pop("OS")
        config_suite = configsuite.ConfigSuite(config, schema)

        self.assertTrue(config_suite.valid)
        self.assertIsNone(config_suite.snapshot.OS)


class TestNotAllowNoneBasicType(unittest.TestCase):
    def test_basictype_explicit_invalid_none(self):
        schema = computers.build_schema()
        config = computers.build_config()

        schema[MK.Content]["OS"][MK.AllowNone] = False
        config["OS"] = None
        config_suite = configsuite.ConfigSuite(config, schema)

        self.assertFalse(config_suite.valid)
        for err in config_suite.errors:
            self.assertTrue(
                isinstance(err, configsuite.validation_errors.InvalidTypeError)
            )
            self.assertTrue(
                err.msg.startswith("Is x a string is false on input 'None'")
            )

        self.assertTrue(config_suite.readable)
        self.assertIsNone(config_suite.snapshot.OS)



"""Copyright 2018 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types

from . import data


class TestSchemaValidation(unittest.TestCase):
    def test_basis_schema(self):
        raw_config = data.candy_bag.build_story_config()
        config_suite = configsuite.ConfigSuite(
            raw_config, data.candy_bag.build_schema()
        )
        self.assertTrue(config_suite.valid)

    def test_missing_type_first_level(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema.pop(MK.Type)
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_extra_key_first_level(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema["unknown"] = "some value"
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_invalid_type_first_level(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema[MK.Required] = "some value"
        with self.assertRaises(TypeError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_dict_missing_content(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema[MK.Content][MK.Item].pop(MK.Content)
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_invalid_schema_dict_value(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema[MK.Content][MK.Item][MK.Content]["color"].pop(MK.Type)
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_invalid_dict_keys(self):
        invalid_keys = ["1monkey", "monkey-donkey", "a b c"]
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()

        for inv_key in invalid_keys:
            schema[MK.Content][MK.Item][MK.Content][inv_key] = {MK.Type: types.String}
            with self.assertRaises(KeyError):
                configsuite.ConfigSuite(raw_config, schema)

    def test_list_missing_content(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema.pop(MK.Content)
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_list_empty_content(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema[MK.Content].pop(MK.Item)
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)

    def test_list_dubious_content(self):
        raw_config = data.candy_bag.build_story_config()
        schema = data.candy_bag.build_schema()
        schema[MK.Content]["dubious"] = "content"
        with self.assertRaises(KeyError):
            configsuite.ConfigSuite(raw_config, schema)



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


@configsuite.validator_msg("Is preference consistence with favourite")
def _preference_favourite_consistency(person, context):
    if person["prefer_normal"]:
        valid_nums = set(context.normal_numbers)
    else:
        valid_nums = set(context.special_numbers)

    favourite_nums = set(person["favourite_numbers"])
    return favourite_nums.issubset(valid_nums)


def extract_context(configuration):
    return configuration


_PERSON_SCHEMA = {
    MK.Type: types.NamedDict,
    MK.ContextValidators: (_preference_favourite_consistency,),
    MK.Content: {
        "prefer_normal": {MK.Type: types.Bool},
        "favourite_numbers": {
            MK.Type: types.List,
            MK.Content: {MK.Item: {MK.Type: types.Integer}},
        },
    },
}


def build_schema():
    return {
        MK.Type: types.NamedDict,
        MK.Content: {
            "special_numbers": {
                MK.Type: types.List,
                MK.Content: {MK.Item: {MK.Type: types.Integer}},
            },
            "normal_numbers": {
                MK.Type: types.List,
                MK.Content: {MK.Item: {MK.Type: types.Integer}},
            },
            "questionnaire": {
                MK.Type: types.List,
                MK.Content: {MK.Item: _PERSON_SCHEMA},
            },
            "mathematicians": {
                MK.Type: types.Dict,
                MK.Content: {MK.Key: {MK.Type: types.String}, MK.Value: _PERSON_SCHEMA},
            },
            "others": {MK.Type: types.NamedDict, MK.Content: {"self": _PERSON_SCHEMA}},
        },
    }


def build_config():
    return {
        "special_numbers": [1, 2, 6, 18],
        "normal_numbers": [3, 4, 5, 7, 15, 31],
        "questionnaire": [
            {"prefer_normal": True, "favourite_numbers": [3, 15]},
            {"prefer_normal": False, "favourite_numbers": [1, 6, 18]},
        ],
        "mathematicians": {
            "Frank P. Ramsey": {
                "prefer_normal": False,
                "favourite_numbers": [1, 2, 6, 18],
            }
        },
        "others": {"self": {"prefer_normal": True, "favourite_numbers": [4, 7, 31]}},
    }



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest

import configsuite

from . import data


class TestLayers(unittest.TestCase):
    def assertEqualSnapshots(self, first, second):
        self.assertEqual(first.heroes, second.heroes)
        self.assertEqual(sorted(first.villains), sorted(second.villains))

    def test_layers_named_dict_disjoint_push(self):
        schema = data.hero.build_schema()

        heroes = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ]
        }

        villains = {"villains": {"Eobard Thawne": 11, "Lux": 3}}

        hero_config = configsuite.ConfigSuite(heroes, schema)
        self.assertFalse(hero_config.valid)
        hero_villains_config = hero_config.push(villains)
        self.assertTrue(hero_villains_config.valid)

        combined = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Eobard Thawne": 11, "Lux": 3},
        }
        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.valid)

        self.assertEqualSnapshots(
            combined_config.snapshot, hero_villains_config.snapshot
        )

    def test_layers_named_dict_intersection_push(self):
        schema = data.hero.build_schema()

        heroes = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Phantom Blot": -10, "Lux": 2},
        }

        villains = {"villains": {"Eobard Thawne": 11, "Lux": 3}}

        hero_config = configsuite.ConfigSuite(heroes, schema)
        self.assertTrue(hero_config.valid)
        hero_villains_config = hero_config.push(villains)
        self.assertTrue(hero_villains_config.valid)

        combined = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Phantom Blot": -10, "Eobard Thawne": 11, "Lux": 3},
        }
        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.valid)

        self.assertEqualSnapshots(
            combined_config.snapshot, hero_villains_config.snapshot
        )

    def test_layers_named_dict_disjoint_init(self):
        schema = data.hero.build_schema()

        heroes = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ]
        }

        villains = {"villains": {"Eobard Thawne": 11, "Lux": 3}}

        layered_config = configsuite.ConfigSuite(heroes, schema, layers=(villains,))
        self.assertTrue(layered_config.valid)

        combined = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Eobard Thawne": 11, "Lux": 3},
        }
        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.valid)

        self.assertEqualSnapshots(combined_config.snapshot, layered_config.snapshot)

    def test_layers_named_dict_intersection_init(self):
        schema = data.hero.build_schema()

        heroes = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Phantom Blot": -10, "Lux": 3},
        }

        layers = ({"villains": {"Eobard Thawne": 11}}, {"villains": {"Lux": 5}})

        layered_config = configsuite.ConfigSuite(heroes, schema, layers=layers)
        self.assertTrue(layered_config.valid)

        combined = {
            "heroes": [
                {"name": "Batman", "strength": 10},
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
            ],
            "villains": {"Eobard Thawne": 11, "Lux": 3, "Phantom Blot": -10},
        }
        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.valid)

        self.assertEqualSnapshots(combined_config.snapshot, layered_config.snapshot)

    def test_multi_layers_dict_intersection(self):
        schema = data.hero.build_schema()

        bottom_layer = {
            "heroes": [],
            "villains": {"Phantom Blot": 0, "Lux": 0, "Joker": 0},
        }
        middle_layer = {"villains": {"Phantom Blot": 1, "Lux": 1}}
        top_layer = {"villains": {"Phantom Blot": 2}}

        layered_config = configsuite.ConfigSuite(
            top_layer, schema, layers=(bottom_layer, middle_layer)
        )
        self.assertTrue(layered_config.valid)

        villains = dict(layered_config.snapshot.villains)
        self.assertEqual(
            sorted(["Phantom Blot", "Lux", "Joker"]), sorted(villains.keys())
        )
        self.assertEqual(0, villains["Joker"])
        self.assertEqual(1, villains["Lux"])
        self.assertEqual(2, villains["Phantom Blot"])

    def test_layers_list_init(self):
        schema = data.hero.build_schema()

        heroes = {
            "heroes": [{"name": "Batman", "strength": 10}],
            "villains": {"Phantom Blot": -10, "Lux": 3, "Eobard Thawne": 11},
        }

        layers = (
            {"heroes": [{"name": "Flash", "strength": 12}]},
            {"heroes": [{"name": "Dirk Gently", "strength": 7}]},
        )

        layered_config = configsuite.ConfigSuite(heroes, schema, layers=layers)
        self.assertTrue(layered_config.valid)

        combined = {
            "heroes": [
                {"name": "Flash", "strength": 12},
                {"name": "Dirk Gently", "strength": 7},
                {"name": "Batman", "strength": 10},
            ],
            "villains": {"Eobard Thawne": 11, "Lux": 3, "Phantom Blot": -10},
        }
        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.valid)

        self.assertEqualSnapshots(combined_config.snapshot, layered_config.snapshot)

    def test_invalid_layer(self):
        schema = data.hero.build_schema()

        dummy_layer = {
            "heroes": [{"name": "Batman", "strength": 10}],
            "villains": {"The joker": "there is not such thing as evil"},
        }

        real_layer = {
            "heroes": [{"name": "Batman", "strength": 10}],
            "villains": {"The Joker": 9},
        }

        layered_config = configsuite.ConfigSuite(
            real_layer, schema, layers=(dummy_layer,)
        )
        self.assertTrue(layered_config.readable)
        self.assertFalse(layered_config.valid)

    def test_invalid_layer_inside_list(self):
        schema = data.hero.build_schema()

        dummy_layer = {"heroes": [{"name": "Batman"}]}

        real_layer = {
            "heroes": [{"name": "Batman", "strength": 10}],
            "villains": {"The Joker": 9},
        }

        layered_config = configsuite.ConfigSuite(
            real_layer, schema, layers=(dummy_layer,)
        )
        self.assertTrue(layered_config.readable)
        self.assertFalse(layered_config.valid)
        self.assertTrue(len(layered_config.errors) > 0)

        combined = {
            "heroes": [{"name": "Batman"}, {"name": "Batman", "strength": 10}],
            "villains": {"The Joker": 9},
        }

        combined_config = configsuite.ConfigSuite(combined, schema)
        self.assertTrue(combined_config.readable)
        self.assertFalse(combined_config.valid)

        self.assertEqualSnapshots(combined_config.snapshot, layered_config.snapshot)



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import collections
import copy
import unittest

import configsuite
from configsuite import MetaKeys as MK
from configsuite import transformation_msg as trans_msg

from .data import templating
from .data import favourite_numbers
from .data import hero


@configsuite.transformation_msg("Put numbers under types in sorted order")
def _sort_favourites(nums_dict):
    nums_dict = copy.deepcopy(nums_dict)
    num_names = ("favourite_uint4", "favourite_uint8", "favourite_int")

    nums = [nums_dict.get(name) for name in num_names]
    neg, non_neg = [x for x in nums if x < 0], [x for x in nums if x >= 0]
    nums = sorted(neg, key=abs) + sorted(non_neg, key=abs)
    nums = collections.deque(nums)

    for name in num_names:
        if name in nums_dict:
            nums_dict[name] = nums.popleft()

    return nums_dict


class TestTransformations(unittest.TestCase):
    def test_transformation_valid(self):
        raw_config = templating.build_config_no_definitions()
        suite = configsuite.ConfigSuite(
            raw_config, templating.build_schema_no_definitions()
        )
        self.assertTrue(suite.valid)
        self.assertEqual(templating.template_results_no_definitions(), suite.snapshot)

    def test_transformation_invalid_value(self):
        raw_config = templating.build_config_no_definitions()
        raw_config.append(14)
        suite = configsuite.ConfigSuite(
            raw_config, templating.build_schema_no_definitions()
        )
        self.assertFalse(suite.valid)

        self.assertEqual(1, len(suite.errors))
        err = suite.errors[0]
        self.assertIsInstance(err, configsuite.InvalidTypeError)

    def test_transformation_exception(self):
        raw_config = templating.build_config_no_definitions()

        schema = templating.build_schema_no_definitions()
        temp_schema = schema[MK.Content][MK.Item]
        _render = temp_schema[MK.Transformation]
        transformation_msg = "Renders Jinja template until the end"
        error_msg = "Did not expect to reach the end!"

        @configsuite.transformation_msg(transformation_msg)
        def _no_end_render(elem):
            if elem == "The end.":
                raise ValueError(error_msg)
            return _render(elem)

        temp_schema[MK.Transformation] = _no_end_render

        suite = configsuite.ConfigSuite(raw_config, schema)
        self.assertFalse(suite.valid)

        self.assertEqual(1, len(suite.errors))
        err = suite.errors[0]
        self.assertIsInstance(err, configsuite.TransformationError)
        error_fmt = "'{}' failed on input '{}' with error '{}'"
        expected_error = error_fmt.format(transformation_msg, "The end.", error_msg)
        self.assertEqual(expected_error, err.msg)
        self.assertEqual((3,), err.key_path)

    def test_transformation_multi_layer(self):
        raw_config = templating.build_config_no_definitions()

        schema = templating.build_schema_no_definitions()
        schema[MK.Transformation] = trans_msg("Max length 3")(lambda l: l[-3:])

        top_layer = ("To be continued.", "... another day.")

        suite = configsuite.ConfigSuite(top_layer, schema)
        self.assertTrue(suite.valid)
        self.assertEqual(top_layer, suite.snapshot)

        layer_suite = configsuite.ConfigSuite(top_layer, schema, layers=(raw_config,))
        self.assertTrue(layer_suite.valid)
        self.assertEqual(
            templating.template_results_no_definitions()[-1:] + top_layer,
            layer_suite.snapshot,
        )

    def test_transformation_list(self):
        raw_config = templating.build_config_no_definitions()

        schema = templating.build_schema_no_definitions()
        schema[MK.Transformation] = trans_msg("Max length 3")(lambda l: l[-3:])

        suite = configsuite.ConfigSuite(raw_config, schema)
        self.assertTrue(suite.valid)
        self.assertEqual(
            templating.template_results_no_definitions()[-3:], suite.snapshot
        )

    def test_transformation_non_readable_results(self):
        raw_config = templating.build_config_no_definitions()

        @configsuite.transformation_msg("Deforming elements")
        def _deformer(elem):
            return {"unexpected_nesting": elem}

        schema = templating.build_schema_no_definitions()
        schema[MK.Transformation] = _deformer

        suite = configsuite.ConfigSuite(raw_config, schema)
        self.assertFalse(suite.readable, suite.errors)

    def test_transformation_nameddict(self):
        schema = favourite_numbers.build_schema()
        schema[MK.Transformation] = _sort_favourites

        raw_config = {
            "favourite_uint4": 1024,
            "favourite_uint8": 7,
            "favourite_int": 42,
        }
        suite = configsuite.ConfigSuite(raw_config, schema)
        self.assertTrue(suite.valid, suite.errors)
        self.assertEqual(7, suite.snapshot.favourite_uint4)
        self.assertEqual(42, suite.snapshot.favourite_uint8)
        self.assertEqual(1024, suite.snapshot.favourite_int)

    def test_transformation_prior_to_validation(self):
        schema = favourite_numbers.build_schema()
        schema[MK.Transformation] = _sort_favourites

        raw_config = {
            "favourite_uint4": 1024,
            "favourite_uint8": 2 ** 4,
            "favourite_int": 42,
        }
        suite = configsuite.ConfigSuite(raw_config, schema)
        self.assertFalse(suite.valid)

    def test_transformation_dict(self):
        @configsuite.transformation_msg("Remove he who must not be named")
        def _remove_unspeakable_names(villains):
            villains = copy.deepcopy(villains)
            if "Voldemort" in villains:
                villains.pop("Voldemort")
            return villains

        schema = hero.build_schema()
        schema[MK.Content]["villains"][MK.Transformation] = _remove_unspeakable_names

        villains = {
            "heroes": [{"name": "Lucky Luke", "strength": 8}],
            "villains": {"The Joker": 7, "Voldemort": 10},
        }
        suite = configsuite.ConfigSuite(villains, schema)
        self.assertTrue(suite.valid, suite.errors)
        self.assertEqual((("The Joker", 7),), suite.snapshot.villains)



"""Copyright 2019 Equinor ASA and The Netherlands Organisation for
Applied Scientific Research TNO.

Licensed under the MIT license.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the conditions stated in the LICENSE file in the project root for
details.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
"""


import unittest


import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


class TestNotAllowNoneVsDefault(unittest.TestCase):
    def test_allow_none_default_not_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: True,
                    MK.Required: False,
                    MK.Default: 0,
                },
            },
        }

        for value in (-1, 4, 1000, None):
            suite = configsuite.ConfigSuite({"my_value": value}, schema)
            self.assertTrue(suite.valid)
            self.assertEqual(value, suite.snapshot.my_value)

        suite = configsuite.ConfigSuite({}, schema)
        self.assertTrue(suite.valid)
        self.assertEqual(0, suite.snapshot.my_value)

    def test_allow_none_no_default_not_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: True,
                    MK.Required: False,
                },
            },
        }

        for value in (-1, 4, 1000, None):
            suite = configsuite.ConfigSuite({"my_value": value}, schema)
            self.assertTrue(suite.valid)
            self.assertEqual(value, suite.snapshot.my_value)

        suite = configsuite.ConfigSuite({}, schema)
        self.assertTrue(suite.valid)
        self.assertEqual(None, suite.snapshot.my_value)

    def test_disallow_none_default_not_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: False,
                    MK.Default: 0,
                    MK.Required: False,
                },
            },
        }

        for value in (-1, 4, 1000):
            suite = configsuite.ConfigSuite({"my_value": value}, schema)
            self.assertTrue(suite.valid, suite.errors)
            self.assertEqual(value, suite.snapshot.my_value)

            suite = configsuite.ConfigSuite({}, schema)
            self.assertTrue(suite.valid)
            self.assertEqual(0, suite.snapshot.my_value)

        suite = configsuite.ConfigSuite({"my_value": None}, schema)
        self.assertFalse(suite.valid)
        self.assertEqual(None, suite.snapshot.my_value)

    def test_disallow_none_no_default_not_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: False,
                    MK.Required: False,
                },
            },
        }

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite({}, schema)
        self.assertIn("A type is not required only if", str(error_context.exception))

    def test_allow_none_default_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: True,
                    MK.Default: 0,
                    MK.Required: True,
                },
            },
        }

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite({}, schema)
        self.assertIn("Required can not have Default", str(error_context.exception))

    def test_allow_none_no_default_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: True,
                    MK.Required: True,
                },
            },
        }

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite({}, schema)
        self.assertIn("A type is not required only if", str(error_context.exception))

    def test_disallow_none_default_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: False,
                    MK.Default: 0,
                    MK.Required: True,
                },
            },
        }

        with self.assertRaises(ValueError) as error_context:
            configsuite.ConfigSuite({}, schema)
        self.assertIn("Required can not have Default", str(error_context.exception))

    def test_disallow_none_no_default_required(self):
        schema = {
            MK.Type: types.NamedDict,
            MK.Content: {
                "my_value": {
                    MK.Type: types.Integer,
                    MK.AllowNone: False,
                    MK.Required: True,
                },
            },
        }

        for value in (-1, 4, 1000):
            suite = configsuite.ConfigSuite({"my_value": value}, schema)
            self.assertTrue(suite.valid)
            self.assertEqual(value, suite.snapshot.my_value)

        for config in ({}, {"my_value": None}):
            suite = configsuite.ConfigSuite(config, schema)
            self.assertFalse(suite.valid)
            self.assertEqual(None, suite.snapshot.my_value)
